/* -*-c-*-
 *
 * $Id: chan.c,v 1.1 1999/07/01 08:56:23 mdw Exp $
 *
 * Channel management
 *
 * (c) 1999 Mark Wooding
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of the `fw' port forwarder.
 *
 * `fw' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * `fw' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with `fw'; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: chan.c,v $
 * Revision 1.1  1999/07/01 08:56:23  mdw
 * Initial revision
 *
 */

/*----- Header files ------------------------------------------------------*/

#include "config.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/uio.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <mLib/alloc.h>
#include <mLib/conn.h>
#include <mLib/sel.h>

#include "chan.h"
#include "fw.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @writechan@ --- *
 *
 * Arguments:	@int fd@ = file descriptor to write to
 *		@unsigned mode@ = what the descriptor is ready for
 *		@void *vp@ = pointer to channel block
 *
 * Returns:	---
 *
 * Use:		Writes to a channel.
 */

static void writechan(int fd, unsigned mode, void *vp)
{
  chan *c = vp;
  int w;
  unsigned base = c->base;
  unsigned len = c->len;

  /* --- Write data from my buffer --- */

  if (len) {

    /* --- Do the write --- */

    if (base + len <= CHAN_BUFSZ)
      w = write(fd, c->buf + base, len);
    else {
      struct iovec iov[2];
      iov[0].iov_base = c->buf + base;
      iov[0].iov_len = CHAN_BUFSZ - base;
      iov[1].iov_base = c->buf;
      iov[1].iov_len = len - iov[0].iov_len;
      w = writev(fd, iov, 2);
    }

    /* --- Sift through the results --- */

    if (w < 0) {
      if (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK)
	return;
      goto close;
    }
    else if (w == 0)
      goto close;
    else if (c->len == CHAN_BUFSZ && !(c->f & CHANF_CLOSE))
      sel_addfile(&c->r);
    c->len -= w;
  }
  if (c->len == 0)
    sel_rmfile(&c->w);

  /* --- Close the output end if necessary --- */

  if (c->len == 0 && (c->f & CHANF_CLOSE)) {
    shutdown(fd, 1);
    c->func(c->p);
  }
  return;

  /* --- Force a close if an error occurred --- */

close:
  chan_close(c);
  c->func(c->p);
}

/* --- @readchan@ --- *
 *
 * Arguments:	@int fd@ = file descriptor to read from
 *		@unsigned mode@ = what the descriptor is ready for
 *		@void *vp@ = pointer to channel block
 *
 * Returns:	---
 *
 * Use:		Reads from a channel.
 */

static void readchan(int fd, unsigned mode, void *vp)
{
  chan *c = vp;
  int r;
  unsigned base = (c->base + c->len) & (CHAN_BUFSZ - 1);
  unsigned len = CHAN_BUFSZ - c->len;

  /* --- Do the read --- */

  if (base + len <= CHAN_BUFSZ)
    r = read(fd, c->buf + base, len);
  else {
    struct iovec iov[2];
    iov[0].iov_base = c->buf + base;
    iov[0].iov_len = CHAN_BUFSZ - base;
    iov[1].iov_base = c->buf;
    iov[1].iov_len = len - iov[0].iov_len;
    r = readv(fd, iov, 2);
  }

  /* --- Sift through the results --- */

  if (r < 0) {
    if (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK)
      return;
    goto close;
  }
  else if (r == 0)
    goto close;
  else if (c->len == 0 && !(c->f & CHANF_READY))
    sel_addfile(&c->w);
  c->len += r;
  if (c->len == CHAN_BUFSZ)
    sel_rmfile(&c->r);
  return;

  /* --- Close the read end of the channel --- */

close:
  c->f |= CHANF_CLOSE;
  if (!c->len)
    sel_addfile(&c->w);
  sel_rmfile(&c->r);
}

/* --- @chan_close@ --- *
 *
 * Arguments:	@chan *c@ = pointer to channel
 *
 * Returns:	---
 *
 * Use:		Closes down a channel prematurely.
 */

void chan_close(chan *c)
{
  if (!(c->f & CHANF_CLOSE) && c->len != CHAN_BUFSZ)
    sel_rmfile(&c->r);
  if ((c->f & CHANF_READY) && c->len != 0)
    sel_rmfile(&c->w);
}

/* --- @chan_dest@ --- *
 *
 * Arguments:	@chan *c@ = pointer to channel
 *		@int fd@ = destination file descriptor for channel
 *
 * Returns:	---
 *
 * Use:		Sets the channel's destination so it knows where to put
 *		data.
 */

void chan_dest(chan *c, int fd)
{
  if (c->f & CHANF_READY)
    return;
  sel_initfile(sel, &c->w, fd, SEL_WRITE, writechan, c);
  if (c->len)
    sel_addfile(&c->w);
}

/* --- @chan_open@ --- *
 *
 * Arguments:	@chan *c@ = pointer to channel to open
 *		@int from, to@ = source and destination file descriptors
 *		@void (*func)(void *p)@ = function to call on closure
 *		@void *p@ = argument to pass to function
 *
 * Returns:	---
 *
 * Use:		Opens a channel.  Data is copied from the source to the
 *		destination.  The @to@ argument may be @-1@ if the file
 *		descriptor isn't known yet.
 */

void chan_open(chan *c, int from, int to,
	       void (*func)(void */*p*/), void *p)
{
  c->func = func;
  c->p = p;

  c->base = 0;
  c->len = 0;
  c->f = 0;

  sel_initfile(sel, &c->r, from, SEL_READ, readchan, c);
  sel_addfile(&c->r);

  if (to != -1)
    chan_dest(c, to);
}

/*----- That's all, folks -------------------------------------------------*/
